/*
Zernike - algorithme de calcul de polynmes de Zernike

Copyright (C) 2002 Observatoire de Paris-Meudon

Ce programme est un logiciel libre ; vous pouvez le redistribuer et/ou le modifier conformment aux dispositions de la Licence Publique Gnrale GNU, telle que publie par la Free Software Foundation ; version 2 de la licence, ou encore ( votre choix) toute version ultrieure.

Ce programme est distribu dans l'espoir qu'il sera utile, mais SANS AUCUNE GARANTIE ; sans mme la garantie implicite de COMMERCIALISATION ou D'ADAPTATION A UN OBJET PARTICULIER. Pour plus de dtail, voir la Licence Publique Gnrale GNU .

Vous devez avoir reu un exemplaire de la Licence Publique Gnrale GNU en mme temps que ce programme ; si ce n'est pas le cas, crivez  la Free Software Foundation Inc., 675 Mass Ave, Cambridge, MA 02139, Etats-Unis.
*/

public class Zernike {

  int dimension;
//  double r;
  double mask;
  double teta;

  public Zernike (int ordre, int _dimension) {
  
    dimension = _dimension;
    double z[][] = new double [dimension][dimension];
//    z = calcZernike (ordre);
  
  }
  
  public static double[][] calcZernike (int zn, int dimension) {

// calcule le zernike normalise (somme(z^2) = 1) de rang zn
// convention ici : zernike(1) = piston, zernike(2) = tilt, etc...

//    System.out.println("On entre dans calcZernike");
    double z[][] = new double[dimension][dimension];
    double tempo;
    double r;
    int znm[] = zernumero(zn);
    int n = znm[0];
    int m = znm[1];
//    System.out.println("On vient d'appeler zernumero. n="+n+"; m="+m);

    if ( (zn % 2) == 1 ) {  // zn odd
      for ( int i=0 ; i <= (n-m)/2 ; i++ ) {
        tempo = Math.pow(-1.,i)*Math.exp(alngam(n-i+1)-(alngam(i+1)+alngam((n+m)/2.-i+1.)+alngam((n-m)/2.-i+1.)));
        for ( int j=0 ; j < dimension ; j++ ) {
          for ( int k=0 ; k < dimension ; k++ ) {
            r = Math.sqrt((j-dimension/2.)*(j-dimension/2.)+(k-dimension/2.)*(k-dimension/2.))/dimension*4.;
            if ( r <= 1. ) {
//            if ( true ) {
              z[j][k]=z[j][k]+Math.pow(r,(n-2.*i))*tempo;
            }
            else {
              z[j][k] = 0.;
            }
          }
        }
      }  
      if ( m == 0 ) {
        for ( int j=0 ; j < dimension ; j++ ) {
          for ( int k=0 ; k < dimension ; k++ ) {
            z[j][k]=z[j][k]*Math.sqrt(n+1.); 
          }
        }
      }
      else {
        for ( int j=0 ; j < dimension ; j++ ) {
          for ( int k=0 ; k < dimension ; k++ ) {
            z[j][k]=z[j][k]*Math.sqrt(2.*(n+1.))*Math.sin(m*Math.atan2((k-dimension/2.),(j-dimension/2.)));
          }
        }
      }
    }
    else {   // zn even
      for ( int i=0 ; i <= (n-m)/2 ; i++ ) {
        tempo = Math.pow(-1.,i)*Math.exp(alngam(n-i+1)-(alngam(i+1)+alngam((n+m)/2.-i+1.)+alngam((n-m)/2.-i+1.)));
        for ( int j=0 ; j < dimension ; j++ ) {
          for ( int k=0 ; k < dimension ; k++ ) {
            r = Math.sqrt((j-dimension/2.)*(j-dimension/2.)+(k-dimension/2.)*(k-dimension/2.))/dimension*4.;
            if ( r <= 1. ) {
//            if ( true ) {
              z[j][k]=z[j][k]+Math.pow(r,(n-2.*i))*tempo;
            }
            else {
              z[j][k] = 0.;
            }
          }
        }
      }
      if ( m == 0 ) {
        for ( int j=0 ; j < dimension ; j++ ) {
          for ( int k=0 ; k < dimension ; k++ ) {
            z[j][k]=z[j][k]*Math.sqrt(n+1.); 
          }
        }
      }
      else {
        for ( int j=0 ; j < dimension ; j++ ) {
          for ( int k=0 ; k < dimension ; k++ ) {
            z[j][k]=z[j][k]*Math.sqrt(2.*(n+1.))*Math.cos(m*Math.atan2((k-dimension/2.),(j-dimension/2.)));
          }
        }
      }
    }

  return z;
  
  }
  
  public static double alngam(double x) {

//**********************************************************************
//
//     DOUBLE PRECISION FUNCTION ALNGAM(X)
//                 double precision LN of the GAMma function
//
//     Returns the natural logarithm of GAMMA(X).
//
//                              Arguments
//     X --> value at which scaled log gamma is to be returned
//                    X is DOUBLE PRECISION
//
//                              Method
//     If X .le. 6.0, then use recursion to get X below 3
//     then apply rational approximation number 5236 of
//     Hart et al, Computer Approximations, John Wiley and
//     Sons, NY, 1968.
//
//     If X .gt. 6.0, then use recursion to get X to at least 12 and
//     then use formula 5423 of the same source.
//
//**********************************************************************
//
//     .. Parameters ..
    double hln2pi=0.91893853320467274178;
//     ..
//     .. Local Scalars ..
    double offset, prod, xx, resu;
    int n;
//     ..
//     .. Local Arrays ..
//    double coef[] = new double[5];
//    double scoefd[] = new double[4];
//    double scoefn[] = new double[9];
//     ..
//     .. Intrinsic Functions ..
//    Intrinsic log,dble,int
//     ..
//     .. Data statements ..
    double scoefn[] = { 0.62003838007127258804E2,
                        0.36036772530024836321E2,
                        0.20782472531792126786E2,
                        0.6338067999387272343E1,
                        0.215994312846059073E1,
                        0.3980671310203570498E0,
                        0.1093115956710439502E0,
                        0.92381945590275995E-2,
                        0.29737866448101651E-2 };
                       
    double scoefd[] = { 0.62003838007126989331E2,
                        0.9822521104713994894E1,
                        -0.8906016659497461257E1,
                        0.1000000000000000000E1 };
                       
    double coef[] = { 0.83333333333333023564E-1,
                      -0.27777777768818808E-2,
                      0.79365006754279E-3,
                      -0.594997310889E-3,
                      0.8065880899E-3 };
//     ..
//     .. Executable Statements ..
    if ( x <= 6.0 ) {
      prod = 1.;
      xx = x;
      if ( x > 3.0 ) {
        while ( xx > 3.0 ) {
          xx = xx - 1.;
          prod = prod*xx;
        }
/*         Do 
            If (.Not. (xx > 3.0E0)) Exit
            xx = xx - one
            prod = prod*xx
         End Do*/
      }
      if (x < 2.0) {
        while ( xx < 2.0 ) {
          prod = prod*xx;
          xx = xx + 1.;
        }
/*         Do 
            If (.Not. (xx < 2.0E0)) Exit
            prod = prod/xx
            xx = xx + one
         End Do*/
      }

      resu = devlpl(scoefn,9,xx-2.0)/devlpl(scoefd,4,xx-2.0);
   //
   //
   //     COMPUTE RATIONAL APPROXIMATION TO GAMMA(X)
   //
   //
      resu = resu*prod;
      resu = Math.log(resu);
      return resu;
      
    }
    else {
      offset = hln2pi;
//
//
//     IF NECESSARY MAKE X AT LEAST 12 AND CARRY CORRECTION IN OFFSET
//
//
      n = (int)(12.0-x);
      if ( n <= 0 ) {
        xx = x;
      }
      else {
        prod = 1.;
        for ( int i = 1 ; i <= n ; i++ ) {
          prod = prod*(x+(double)(i-1));
        }
        offset = offset - Math.log(prod);
        xx = x + (double)(n);
      }
   //
   //
   //     COMPUTE POWER SERIES
   //
   //
      resu = devlpl(coef,5,1./Math.pow(xx,2.))/xx;
      resu = resu + offset + (xx-0.5)*Math.log(xx) - xx;
      return resu;
      
    }
  }

  public static double devlpl(double a[], int n, double x) {

//**********************************************************************
//
//     Double Precision FUNCTION DEVLPL(A,N,X)
//              Double precision EVALuate a PoLynomial at X
//     returns
//          A(0) + A(1)*X + ... + A(N-1)*X**(N-1)
//
//
//                              Arguments
//     A --> Array of coefficients of the polynomial.
//                                        A is Double Precision(N)
//
//     N --> Length of A, also degree of polynomial - 1.
//                                        N is INTEGER
//
//     X --> Point at which the polynomial is to be evaluated.
//                                        X is Double Precision
//
//**********************************************************************
//
//     .. Local Scalars ..
    double term;
//     .. Executable Statements ..
    term = a[n-1];
    for ( int i = n - 2 ; i >= 0 ; i-- ) {
       term = a[i] + term*x;
    }
    
    return term;
  
  }

  public static void complexToComplex2D(int sign, int n, double ar[][], double ai[][]) {
  
    double tempoRe[] = new double[n];
    double tempoIm[] = new double[n];
    
    for ( int i = 0 ; i < n ; i++ ) {
      for ( int j = 0 ; j < n ; j++ ) {
        tempoRe[j] = ar[i][j];
        tempoIm[j] = ai[i][j];
      }
      complexToComplex(sign, n, tempoRe, tempoIm);
      for ( int j = 0 ; j < n ; j++ ) {
        ar[i][j] = tempoRe[j];
        ai[i][j] = tempoIm[j];
      }
    }
    for ( int j = 0 ; j < n ; j++ ) {
      for ( int i = 0 ; i < n ; i++ ) {
        tempoRe[i] = ar[i][j];
        tempoIm[i] = ai[i][j];
      }
      complexToComplex(sign, n, tempoRe, tempoIm);
      for ( int i = 0 ; i < n ; i++ ) {
        ar[i][j] = tempoRe[i];
        ai[i][j] = tempoIm[i];
      }
    }
  }

  public static void complexToComplex(int sign, int n, double ar[], double ai[]) {
    double scale = Math.sqrt(1.0/n);

    int i,j;
    for (i=j=0; i<n; ++i) {
      if (j>=i) {
        double tempr = ar[j]*scale;
	double tempi = ai[j]*scale;
	ar[j] = ar[i]*scale;
	ai[j] = ai[i]*scale;
	ar[i] = tempr;
	ai[i] = tempi;
      }
      int m = n/2;
      while (m>=1 && j>=m) {
        j -= m;
        m /= 2;
      }
      j += m;
    }
    
    int mmax,istep;
    for (mmax=1,istep=2*mmax; mmax<n; mmax=istep,istep=2*mmax) {
      double delta = sign*3.141592654/(double)mmax;
      for (int m=0; m<mmax; ++m) {
        double w = (double)m*delta;
        double wr = Math.cos(w);
        double wi = Math.sin(w);
        for (i=m; i<n; i+=istep) {
          j = i+mmax;
          double tr = wr*ar[j]-wi*ai[j];
          double ti = wr*ai[j]+wi*ar[j];
          ar[j] = ar[i]-tr;
          ai[j] = ai[i]-ti;
          ar[i] += tr;
          ai[i] += ti;
        }
      }
      mmax = istep;
    }
  }

  public static int[] zernumero(int zn) {

// calcul de l'ordre radial et azimutal d'un zernike donne
// retourne un vecteur a deux valeurs : ordre radial, puis azimutal

    int j = 0;
    int resu[] = { 0 , 0 };
    
    for ( int n=0 ; n <=100 ; n++ ) {    
      for ( int m=0 ; m <= n ; m++ ) {      
	if ( (n-m) % 2 == 0 ) { // n-m even
          j=j+1;
          if ( j == zn ) {
            resu[0] = n;
            resu[1] = m;
            break;
          }
          if ( m != 0 ) {
            j=j+1;
            if ( j == zn ) {
              resu[0] = n;
              resu[1] = m;
              break;
            }
          }
        } 
      }
    }
    return resu;
  }
  
}
