/*
Speckle - algorithme de simulation pour SimuApplet

Copyright (C) 2002 Observatoire de Paris-Meudon

Ce programme est un logiciel libre ; vous pouvez le redistribuer et/ou le modifier conformment aux dispositions de la Licence Publique Gnrale GNU, telle que publie par la Free Software Foundation ; version 2 de la licence, ou encore ( votre choix) toute version ultrieure.

Ce programme est distribu dans l'espoir qu'il sera utile, mais SANS AUCUNE GARANTIE ; sans mme la garantie implicite de COMMERCIALISATION ou D'ADAPTATION A UN OBJET PARTICULIER. Pour plus de dtail, voir la Licence Publique Gnrale GNU .

Vous devez avoir reu un exemplaire de la Licence Publique Gnrale GNU en mme temps que ce programme ; si ce n'est pas le cas, crivez  la Free Software Foundation Inc., 675 Mass Ave, Cambridge, MA 02139, Etats-Unis.
*/

import java.util.Vector;

import simu.*;

/**
 * Algorithme de simulation pour SimuApplet.
 * Dformation d'une image par la turbulence atmosphrique.
 */
public class Speckle extends SimuApplet {

    double tmasque[][];
    int dx=256, dy=128, dz;
    double z0[][];
    double w;
//    int tps;
    double champ[][];
    
    int mz;            // nombre de Zernikes
    double zi[][][];   // tableau pour mzZernikes
    double ph[][];     // phase instantane
    double z[];        // coefficients des Zernikes
    double tps[];      // enveloppe stat. de Kolmogorov
    double maxph;      // maximum de la phase (pour affichage)
    double maxi;       // maximum de l'intensit (pour affichage)
    double pupille[][];// pupille unit

    public Speckle(){
    
/*    int n[] = new int[2];   //  petit exo pour verifier zernumero
    for ( int zn = 1 ; zn < 7 ; zn++ ) {
      n = Zernike.zernumero(zn);
      System.out.println("zn="+zn+"; n="+n[0]+"; m="+n[1]);
    }*/
    
        ParamIn v = new ParamIn();
        v.type = "nombre";
        v.label = "doverR";
        v.titre = "D/r0";
        v.unit = "sans unit";
        v.acquisition = "champ";
        v.defaut = "30";
        addParamIn(v);
        
        ParamIn l = new ParamIn();
        l.type = "nombre";
        l.label = "nb_zernike";
        l.titre = "Nombre de polynmes de Zernike cumuls";
        l.acquisition = "champ";
        l.defaut = "20";
        addParamIn(l);
        
        ParamIn z2 = new ParamIn();
        z2.type = "string";
        z2.label = "z2";
        z2.titre = "Zernike ordre 2 (Tip)";
        z2.acquisition = "case";
        z2.defaut = "on";
        addParamIn(z2);
        
        ParamIn z3 = new ParamIn();
        z3.type = "string";
        z3.label = "z3";
        z3.titre = "Zernike ordre 3 (Tilt)";
        z3.acquisition = "case";
        z3.defaut = "on";
        addParamIn(z3);
        
        ParamIn z4 = new ParamIn();
        z4.type = "string";
        z4.label = "z4";
        z4.titre = "Zernike ordre 4 (Focus)";
        z4.acquisition = "case";
        z4.defaut = "on";
        addParamIn(z4);
        
        ParamIn z5 = new ParamIn();
        z5.type = "string";
        z5.label = "z5";
        z5.titre = "Zernike ordre 5 (Astigmatisme 0)";
        z5.acquisition = "case";
        z5.defaut = "on";
        addParamIn(z5);
        
        ParamIn z6 = new ParamIn();
        z6.type = "string";
        z6.label = "z6";
        z6.titre = "Zernike ordre 6 (Astigmatisme 45)";
        z6.acquisition = "case";
        z6.defaut = "on";
        addParamIn(z6);
        
        ParamIn z7etPlus = new ParamIn();
        z7etPlus.type = "string";
        z7etPlus.label = "z7etPlus";
        z7etPlus.titre = "Zernike ordres 7 et + (Comas et suivants)";
        z7etPlus.acquisition = "case";
        z7etPlus.defaut = "on";
        addParamIn(z7etPlus);
        
        // paramtres de sortie
        outparams = new Vector();
        ParamOut chp = new ParamOut();
        chp.type = "tableau";
        chp.label = "champ";
        addParamOut(chp);

        // options d'affichage
        affichage = new Vector();
        Affichage p = new Affichage();
        p.type = "image";
        p.titre = "Lancer le calcul";
        String[][]pparams = {{"champ"}};
        p.params = pparams;
        p.imgdimx = dx;
        p.imgdimy = dy;
        addAffichage(p);
    appletDoc("speckledoc.xml");

  }


// calcul en image (utilis avec affichage image)
    public void initCalculLive(ListeValeurs in) throws SimuException {

// on rcupre le nb de Zernikes
      mz = (int) in.lireDouble("nb_zernike");
// on rcupre les infos on/off sur les Zernikes
      String z2 = in.lireString("z2");
      String z3 = in.lireString("z3");
      String z4 = in.lireString("z4");
      String z5 = in.lireString("z5");
      String z6 = in.lireString("z6");
      String z7etPlus = in.lireString("z7etPlus");
        
      ph = new double[dx/2][dy];
      if ( mz <= 1 ) mz = 1;
      z = new double[mz];
      zi = new double[dx/2][dy][mz];
      tps = new double[mz];
      double som = 0.;
      int zn[] = { 0 , 0 };
      double amp;
      double table_temp[][] = new double[dx/2][dy];
      double table_nulle[][] = new double[dx/2][dy];
      double toggle[] = new double[mz];
      double doverR = in.lireDouble("doverR");
        
// on allume ou teint les Zernikes

      for ( int i = 0 ; i < mz ; i++ ) {
        toggle[i] = 1.;      
      }        
      
      if ( mz > 0 && z2 == "off" ) toggle[0] = 0.;
      if ( mz > 1 && z3 == "off" ) toggle[1] = 0.;
      if ( mz > 2 && z4 == "off" ) toggle[2] = 0.;
      if ( mz > 3 && z5 == "off" ) toggle[3] = 0.;
      if ( mz > 4 && z6 == "off" ) toggle[4] = 0.;
      if ( mz > 5 && z7etPlus == "off" ) {
        for ( int i = 5 ; i < mz ; i++ ) {
          toggle[i] = 0;      
        }        
      }
        
// rcupration d'une pupille unit

      pupille = Zernike.calcZernike(1,dx/2);
        
// calcul et stockage de mz zernikes avec spectre de K en -11/3 en variance et -11/6 en amplitude
      for ( int o=0 ; o <= mz-1 ; o++ ) {
	zn = Zernike.zernumero(o + 2);
        
        table_temp = Zernike.calcZernike(o+2,dx/2);
        
        for(int i=0; i<dx/2; i++){
          for(int j=0; j<dy; j++){
	    zi[i][j][o] = table_temp[i][j];
          }
        }
        
    	tps[o]    = toggle[o] * 1.0 / Math.pow(zn[0],(11.0 / 6.0));
	som       = som + tps[o] * tps[o];
      }
      
// normalisation pour doverR
      amp = Math.sqrt(1.03*Math.pow(doverR,(5.0/3.0))/som);
      for ( int i = 0 ; i < mz ; i++ ) {
        tps[i] = tps[i] * amp;
      }
      
// calcul des parametres de display optimaux
// si tous les zernikes sont a 1, que vaut la phase ?
      maxph = 0.;
      for ( int i = 0 ; i < mz ; i++ ) {
        maxph = maxph + tps[i];
      }
      
      table_temp = Zernike.calcZernike(1,dx/2);
      Zernike.complexToComplex2D(1,dx/2,table_temp,table_nulle);
      
      for(int i=0; i<dx/2; i++){
        for(int j=0; j<dy; j++){
	  maxi = Math.max(maxi,(table_temp[i][j]*table_temp[i][j]));
        }
      }
      maxi = maxi * 10./doverR/doverR;
      
      champ = new double[dx][dy];
    }
    
// renvoit le double  l'index "index" dans le Vector "v"
    public static double getDouble(Vector v, int index) throws SimuException {
        if (index < 0 || index >= v.size())
            throw new SimuException("getDouble: index trop grand");
        double d;
        String sd = (String)v.elementAt(index);
        d = (new Double(sd)).doubleValue();
        return(d);
    }
    
// renvoit le double  l'index "index" dans le Vector "v"
    public static String getString(Vector v, int index) throws SimuException {
        if (index < 0 || index >= v.size())
            throw new SimuException("getDouble: index trop grand");
        String sd = (String)v.elementAt(index);
        return(sd);
    }
    
    public ListeValeurs calculLive() throws SimuException {

      ListeValeurs out = new ListeValeurs();
      double ph[][] = new double[dx/2][dy]; 
      double eiphiRe[][] = new double[dx/2][dy];
      double eiphiIm[][] = new double[dx/2][dy];
      double ims[][] = new double[dx/2][dy];

// on gnre l'intensit des modes
      for ( int i = 0 ; i < mz ; i++ ) {
        z[i] = tps[i] * 2. * (Math.random()-0.5);
      }
      
// on cre la phase instantane
      for ( int o=0 ; o <= mz-1 ; o++ ) {
        for(int i=0; i<dx/2; i++) {
          for(int j=0; j<dy; j++) {
            ph[i][j] = ph[i][j] + z[o] * zi[i][j][o];
          }
        }
      }
      
// on cre l'amplitude complexe, avec un signe - un point sur 2 afin d'avoir un rltat de TF centr
    for(int i=0; i<dx/2; i++) {
        for(int j=0; j<dy; j++) {
          eiphiRe[i][j] = Math.pow(-1.,(double)i)*Math.pow(-1.,(double)j)*Math.cos(ph[i][j])*pupille[i][j];
          eiphiIm[i][j] = Math.pow(-1.,(double)i)*Math.pow(-1.,(double)j)*Math.sin(ph[i][j])*pupille[i][j];
        }
      }
      
// on calcule la TF de l'amplitude complexe
      Zernike.complexToComplex2D(1,dx/2,eiphiRe,eiphiIm);
      
// on calcule l'intensit
      for(int i=0; i<dx/2; i++) {
        for(int j=0; j<dy; j++) {
          ims[i][j] = eiphiRe[i][j]*eiphiRe[i][j]+eiphiIm[i][j]*eiphiIm[i][j];
        }
      }
      
      for(int i=0; i<dx; i++) {
        for(int j=0; j<dy; j++) {
          if ( i<dx/2 ) 
            champ[i][j] = ph[i][j];
          else
            champ[i][j] = ims[i-dx/2][j];
        }
      }

        for(int i=0; i<dx; i++) {
            for(int j=0; j<dy; j++) {
              if ( i<dx/2 )
                champ[i][j] = 255.*champ[i][j]/maxph+128;
              else
                champ[i][j] = 255.*champ[i][j]/maxi;
              if ( champ[i][j] > 255 ) champ[i][j] = 255;
              if ( champ[i][j] < 0 ) champ[i][j] = 0;
            }
        }
//        champ = pupille;
        out.ajouter("champ",champ);
//        System.out.println("maximage="+maximage);        
//        System.out.println("minimage="+minimage);        
//        System.out.println("tps="+tps);        
        return out;
    }
    
    public ListeValeurs calcul(ListeValeurs in) throws SimuException {
        throw new SimuException("calcul n'est pas implment dans Onde: " +
            "utiliser calculLive  la place");
    }

}
